/* This is "heat10.c", written to accompany Introduction to
   Scientific Programming by Joseph L. Zachary. */

/* This program is an implementation and test driver for the
   heat transfer problem described in Chapter 17 of the text.
   It produces Mathematica animation output as described in
   Chapter 18 of the text.  It is not quite correct, because
   it manages file output incorrectly. */


#include <stdio.h>
#include <stdlib.h>

#define MAXSIZE 1001    /* The maximum number of segments that can be */
                        /* simulated.  This is all that must be modified */
                        /* to change the maximum size of the simulation. */



/* Returns the largest of "x", "y", and "z". */

double max (double x, double y, double z)
{
  if (x >= y && x >= z) {
    return(x);
  }
  if (y >= z) {
    return(y);
  }
  return(z);
}



/* Returns the smallest of "x", "y", and "z". */

double min (double x, double y, double z)
{
  if (x <= y && x <= z) {
    return(x);
  }
  if (y <= z) {
    return(y);
  }
  return(z);
}



/* This function initializes "cur" to reflect the segment
   boundary temperatures at time zero.  There are "segs" segments.
   The temperature of the "left" heat source, the "right" heat source,
   and the "rod" are given. */

void initRod (double cur[], int segs,
              double left, double right, double rod)
{
  int i;                /* Steps through "cur". */

  cur[0] = left;
  for (i = 1; i < segs; i = i+1) {
    cur[i] = rod;
  }
  cur[segs] = right;
}



/* This function modifies "cur" to reflect the passage of one tick
   of simulated time.  There are "segs" segments. */

void nextRod (double cur[], int segs)
{

  double nxt[MAXSIZE];  /* Used to temporarily hold the new temperatures */
                        /* of "cur" during the simulation. */
  
  int i;                /* Steps through "cur" and "nxt". */

  for (i = 1; i < segs; i = i+1) {
    nxt[i] = (cur[i-1] + cur[i+1]) / 2;
  }
  
  for (i = 1; i < segs; i = i+1) {
    cur[i] = nxt[i];
  }

}  



/* This function writes to a file a Mathematica command to create
   a single frame of an animation showing the temperatures in "rod",
   which has the specified number of "segments".  The "length" of the
   rod in centimeters, as well as the "low" and "high" temperature
   bounds of the vertical axis are also given. */

void createFrame (double rod[], int segments,
		  double length, double low, double high)
{

  int i;            /* Steps through "rod" */
  FILE *outFile;    /* Pointer to output file */


  /* Obtain a file pointer to the file "heat-output.txt".  If the
     file cannot be opened for writing, print an explanatory
     message and exit the program immediately. */

  outFile = fopen("heat-output.txt", "w");
  if (!outFile) {
    printf("Unable to open file heat-output.txt\n");
    exit(1);
  }


  /* Write to "heat-output.txt" the Mathematica command that
     creates the frame. */

  fprintf(outFile, "ListPlot[{");

  for (i = 0; i < segments; i = i+1) {
    fprintf(outFile, "{%g,%g},",
	    (i*length/segments), rod[i]);
  }

  fprintf(outFile, "{%g,%g}}, ", length, rod[segments]);
  fprintf(outFile, "PlotJoined->True, PlotRange->{%g,%g}]\n\n",
          low, high);

  fclose(outFile);

  
  /* Close "heat-output.txt". */

  fclose(outFile);

}



/* This function models heat transfer in a rod of uniform temperature
   that is insulated along its lengths and has a heat source applied
   to its ends.  The rod's thermal "diffusivity" in cm^2/sec and
   "length" in cm are given, as are the temperature of the "rod", the
   "left" heat source, and the "right" heat source in degrees Celsius.

   The function returns the approximate temperature of the center of
   the rod in degrees Celsius "duration" seconds after the heat sources
   are applied.  It also produces an animation of the heat transfer
   process.

   The implementation is based on a finite-element model with "segs"
   elements. */

double simulate (double diffusivity, double length,
                 double left, double right, double rod,
                 double duration, int segs)
{

  double delta,         /* Length in seconds of each simulated time step */

         clock,         /* Repeatedly counts simulated time from zero to */
                        /* one second before being reset. */

         low,           /* The lowest and highest of the right end, left */
	 high;		/* end, and inititial rod temperatures. */

  double cur[MAXSIZE];  /* Temperatures at the ends of each of the "segs" */
                        /* segments, where cur[0] is the temperature of */
                        /* the left heat source and cur[segs+1] is the */
                        /* temperature of the right heat source. */
  

  /* Initialize  "cur" and compute the length of the simulated
     time step. */

  initRod(cur, segs, left, right, rod);
  delta = (length*length/(segs*segs)) / (2*diffusivity);


  /* Determine the minimum and maximum temperatures in the simulation. */

  low = min(rod, left, right);
  high = max(rod, left, right);


  /* Create a frame of the animation corresponding time zero. */

  createFrame(cur, segs, length, low, high);


  /* Set the one-second timer to zero. */

  clock = 0;


  /* Simulate the passage of time.  During each iteration, compute the
     new segment boundary temperatures in terms of the old segment
     boundary temperatures.  As each second passes, create a frame
     of the animation. */

  while (duration > 0) {

    nextRod(cur, segs);
    duration = duration - delta;

    clock = clock + delta;
    if (clock >= 1 - delta/2) {
      createFrame(cur, segs, length, low, high);
      clock = clock - 1;
    }
      
  }


  /* Return the temperature of the center segment boundary. */  

  return(cur[segs/2]);

}



/* Test driver for "simulate". */

void main (void)
{

  double C, length, left,   /* Constants specified by the user. */
         right, rod, time,
         temp;

  int segs;                 /* Size of finite-element simulation. */

  FILE *inFile;             /* Pointer to input file. */


  /* Obtain a file pointer to the file "heat-input.txt".  If the
     file cannot be opened for reading, print an explanatory
     message and exit the program immediately. */

  inFile = fopen("heat-input.txt", "r");
  if (!inFile) {
      printf("Could not open heat-input.txt\n");
      exit(1);
  }


  /* Read constants from "heat-input.txt" using "fscanf". */

  fscanf(inFile, "%lf", &C);
  fscanf(inFile, "%lf", &length);
  fscanf(inFile, "%lf%lf%lf", &left, &rod, &right);
  fscanf(inFile, "%lf", &time);
  fscanf(inFile, "%d", &segs);


  /* Close "heat-input.txt". */

  fclose(inFile);


  /* Calculate and display the result. */

  temp = simulate(C, length, left, right, rod, time, segs);
  printf("Temperature at center is %g\n", temp);

}
