(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     12633,        487]*)
(*NotebookOutlinePosition[     13286,        510]*)
(*  CellTagsIndexPosition[     13242,        506]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell[TextData[{
  "This notebook contains the Mathematica commands from Appendix A of ",
  StyleBox["Introduction to Scientific Programming",
    FontSlant->"Italic"],
  " by Joseph L. Zachary."
}], "Subsection"],

Cell[TextData[{
  "(A.1)  An example of how ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " can automatically perform dimensional analysis."
}], "Text"],

Cell["72 in * 2.54 cm/in", "Input"],

Cell[TextData[{
  "(A.2)  The expression from (A.1) is equivalent to this one.  This \
illustrates how, in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ", the multiplication operator can be omitted."
}], "Text"],

Cell["72 * in * 2.54 * cm / in", "Input"],

Cell[TextData[{
  "(A.3)  Here we see that ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " does dimensional analysis by combining and cancelling symbolic constants. \
 It doesn't know that one foot is 12 inches."
}], "Text"],

Cell["6 ft * 2.54 cm/in", "Input"],

Cell[TextData[{
  "(A.4)   If, however, we first tell ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " what a foot is ..."
}], "Text"],

Cell[BoxData[
    \(ft\  = \ 12\ in\)], "Input"],

Cell["(A.5)  ... we obtain the simplification that we expected.", "Text"],

Cell[BoxData[
    \(\(\n6\ ft\ *\ 2.54\ cm/in\)\)], "Input"],

Cell["Let's clear the definition of \"ft\".", "Text"],

Cell[BoxData[
    \(Clear[ft]\)], "Input"],

Cell[TextData[{
  "(A..6)  When doing dimensional analysis, you must be careful to keep ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "'s precedence rules in mind.   This expression doesn't simplify properly \
..."
}], "Text"],

Cell["360 deg / 7.2 deg * 5000. stade * 0.1575 km/stade", "Input"],

Cell["\<\
(A.7)   ... until we group \"7.2\" and \"deg\" together with \
parentheses.\
\>", "Text"],

Cell["360 deg / (7.2 deg) * 5000. stade * 0.1575 km/stade", "Input"],

Cell[TextData[{
  "(A.8)  Here we see that ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " doesn't always simplify units as much as might seem reasonable.  The \
problem here is that there are two possible ways to simplify the square \
root."
}], "Text"],

Cell["Sqrt[(2.09*^7 ft + 66 ft)^2 - (2.09*^7 ft)^2]", "Input"],

Cell["\<\
(A.9)  This is the definition of \"horizontal\" from Chapter \
7,\
\>", "Text"],

Cell["horizontal[v_, theta_, t_] := v * t * Cos[theta]", "Input"],

Cell["(A.10)  this is the definition of \"vertical\",", "Text"],

Cell["\<\
vertical[v_, theta_, t_] := v * t * Sin[theta] - 1/2 * g * t^2\
\>",
   "Input"],

Cell["\<\
(A.11)  and this is the gravitational constant, which we have \
defined with appropriate units.\
\>", "Text"],

Cell["g = 9.8 m/sec^2", "Input"],

Cell["\<\
(A.12)  The call to \"horizontal\" yields a properly dimensioned \
result,\
\>", "Text"],

Cell["horizontal[100 m/sec, Pi/4., 10 sec]", "Input"],

Cell["(A.13)  as does this call to \"vertical\".", "Text"],

Cell["vertical[100 m/sec, Pi/4., 10 sec]", "Input"],

Cell["\<\
(A.14)  Adding a unit to Pi/4 doesn't work, however, because the \
built-in \"Cos\" function isn't expecting it.\
\>", "Text"],

Cell["horizontal[100 m/sec, Pi/4. rad, 10 sec]", "Input"],

Cell["\<\
(A.15)  Units don't mix well with plotting commands either, as both \
this example ...\
\>", "Text"],

Cell["\<\
Plot[vertical[100 m/sec, Pi/4, t],
            {t, 0 sec, 20 sec}]\
\>", "Input"],

Cell["(A.16)  ... and this example illustrate.", "Text"],

Cell["Plot[vertical[100, Pi/4, t], {t, 0, 20}]", "Input"],

Cell[TextData[{
  "(A.17)  ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "'s \"Units\" package provides help with unit conversions."
}], "Text"],

Cell[BoxData[
    \(Needs["\<Miscellaneous`Units`\>"]\)], "Input"],

Cell["\<\
(A.18)  The \"Convert\" function requires that you spell the names \
of the units exactly as it expects.\
\>", "Text"],

Cell[BoxData[
    \(Convert[60\ Mile/Hour, \ Kilo\ Meter/Second]\)], "Input"],

Cell[TextData[{
  "(A.19)  Here is an example of how calculations can be typeset in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "."
}], "Text"],

Cell[BoxData[
    \(rat10 = 
      1\/2 + 1\/4 + 1\/6 + 1\/8 + 1\/10 + 1\/12 + 1\/14 + 1\/16 + 1\/18 + 
        1\/20\)], "Input"],

Cell["\<\
(A.20)  Here is an example involving a radical and an \
exponent.\
\>", "Text"],

Cell[BoxData[
    \(\@\(2*2.09*^7*100 + 100\^2\)\)], "Input"],

Cell["\<\
(A.21)  Typesetting can be used when defining functions, as \
illustrated here ...\
\>", "Text"],

Cell[BoxData[
    \(horizon[R_, h_]\  := \ \@\(2  Rh + h\^2\)\)], "Input"],

Cell["(A.22)  ... and here.", "Text"],

Cell[BoxData[
    \(compound[p_, r_, n_, m_]\  := \ p \((1 + r\/m)\)\^\(m\ n\)\)], "Input"],

Cell["(A.23)  Here is one example of typesetting a Greek letter,", "Text"],

Cell[BoxData[
    \(horizontal[100, \ \[Pi]\/4, \ 10]\)], "Input"],

Cell["(A.24)  and here is another.", "Text"],

Cell[BoxData[
    \(horizontal[v_, \ \[Theta]_, \ t_]\  := \ v\ t\ Cos[\[Theta]]\)], "Input"],

Cell["\<\
(A.25)  Here is one example of a typeset Calculus expression \
...\
\>", "Text"],

Cell[BoxData[
    \(makePower[pos_, \ mass_]\  := \ 
      mass\ \[PartialD]\_t\ \(\[PartialD]\_t\ pos\)\ \[PartialD]\_t\ pos\)], 
  "Input"],

Cell["(A.26)  ... and here is another.", "Text"],

Cell[BoxData[
    \(\[Integral]\_0.0\%8.0\((Sin[x] + 2)\) \[DifferentialD]x\)], "Input"],

Cell["\<\
(A.27)  The \"ComputerArithmetic\" package will allow you to \
experiment with simulated floating-point number systems.\
\>", "Text"],

Cell["Needs[\"NumericalMath`ComputerArithmetic`\"]", "Input"],

Cell["\<\
(A.28)  We specify trhat we wish to experiment with a base-ten \
system with two-digit mantissas and exponents that range between -1 and 1, \
inclusive.\
\>", "Text"],

Cell["SetArithmetic[2, 10, ExponentRange->{-1, 2}]", "Input"],

Cell["(A.29)  In this system, 11.2 is represented by 11,", "Text"],

Cell["ComputerNumber[11.2]", "Input"],

Cell["(A.30)  0.555 is represented by 0.56,", "Text"],

Cell["ComputerNumber[.555]", "Input"],

Cell["(A.31)  112. causes an overflow, and", "Text"],

Cell["ComputerNumber[112.]", "Input"],

Cell["(A.32)  0.0555 causes an underflow.", "Text"],

Cell["ComputerNumber[.0555]", "Input"],

Cell["\<\
(A.33)  We can do arithmetic in the simulated system.  Here we find \
the two-digit result of dividing 1 by 8 ...\
\>", "Text"],

Cell["ComputerNumber[1] / ComputerNumber[8]", "Input"],

Cell["\<\
(A.34)  ... and here we find the two-digit result of multiplying 12 \
and 1.2.\
\>", "Text"],

Cell["ComputerNumber[12] * ComputerNumber[1.2]", "Input"],

Cell["\<\
(A.35)  We can change to a new simulated system at any time.  Here, \
we specify that the base is ten, that there are two digits of mantissa, and \
that the exponents range from -2 to 2.\
\>", "Text"],

Cell["SetArithmetic[2, 10, ExponentRange->{-2, 3}]", "Input"],

Cell["\<\
(A.36)  The next two expressions demonstrate that multiplication is \
not associative ...\
\>", "Text"],

Cell["\<\
quot1 = (ComputerNumber[14] * ComputerNumber[11]) * \
ComputerNumber[2]\
\>", "Input"],

Cell["(A.37)  ... in a floating-point system.", "Text"],

Cell["\<\
quot2 = ComputerNumber[14] *
         (ComputerNumber[11] * ComputerNumber[2])\
\>", "Input"],

Cell["\<\
(A.38)  We can use \"InputForm\" to learn more about a simulated \
floating-point number.\
\>", "Text"],

Cell["InputForm[quot1]", "Input"],

Cell["(A.39)  Compare this number with the one from (A.38).", "Text"],

Cell["InputForm[quot2]", "Input"],

Cell["\<\
(A.40)  \"Arithmetic\" shows us the details of the floating-point \
system currently being simulated.\
\>", "Text"],

Cell["Arithmetic[]", "Input"],

Cell["(A.41)  This is a 30-digit arbitrary-precision number.", "Text"],

Cell["radius = N[1/7, 30]", "Input"],

Cell["(A.42)  This is a 25-digit approximation of Pi.", "Text"],

Cell["pi = N[Pi, 25]", "Input"],

Cell["(A.43)  This is a 25-digit approximation of Pi/49.", "Text"],

Cell["area = N[Pi*(1/7)^2, 25]", "Input"],

Cell["\<\
(A.44)  Arbitrary-precision numbers keep track of how many digits \
are meaningful, as illustrated here,\
\>", "Text"],

Cell["Precision[radius]", "Input"],

Cell["(A.45) and here,", "Text"],

Cell["Precision[pi]", "Input"],

Cell["(A.46)  and here.", "Text"],

Cell["Precision[area]", "Input"],

Cell["\<\
(A.47)  We calculate a number using a 30-digit radius and a \
25-digit pi.\
\>", "Text"],

Cell["area = pi * radius^2", "Input"],

Cell["(A.48)  The result has 25 digits.", "Text"],

Cell["Precision[area]", "Input"],

Cell["\<\
(A.49)  Squaring a 25-digit number produces a 25-digit number.\
\>",
   "Text"],

Cell["Precision[pi^2]", "Input"],

Cell["\<\
(A.50)  Digits are gradually lost as more operations are performed.\
\
\>", "Text"],

Cell["Precision[pi^4]", "Input"],

Cell["\<\
(A.51)  Here, we have lost five digits from the original 25-digit \
precision of pi.\
\>", "Text"],

Cell["Precision[pi^100000]", "Input"],

Cell["\<\
(A.52)  When converting a number into arbitrary-precision form, \
always specify a rational number.  Otherwise, the result will have only 16 \
digits of precision.\
\>", "Text"],

Cell["Precision[N[1.234, 50]]", "Input"],

Cell["(A.53)  Writing an extra-long \"floating-point\" number ...", "Text"],

Cell["x = .12345678901234567890*^20", "Input"],

Cell["\<\
(A.54)  ...  is another way to create an arbitrary-precision \
number.\
\>", "Text"],

Cell["Precision[x]", "Input"],

Cell["(A.55)  The largest", "Text"],

Cell["$MaxNumber", "Input"],

Cell["(A.56)  and smallest arbitrary-precision numbers,", "Text"],

Cell["$MinNumber", "Input"],

Cell["(A.57)  are much more extreme than the largest", "Text"],

Cell["$MaxMachineNumber", "Input"],

Cell["(A.58)  and smallest floating-point numbers.", "Text"],

Cell["$MinMachineNumber", "Input"],

Cell["\<\
(A.59)  Instead of overflowing, large floating-point results are \
converted into arbitrary-precision numbers ...\
\>", "Text"],

Cell["$MaxMachineNumber^2", "Input"],

Cell["(A.60)  ... unless the result is too big.", "Text"],

Cell["$MaxMachineNumber ^ $MaxMachineNumber", "Input"],

Cell[TextData[{
  "(A.62)  Once you have translated heat12.tm into a C program, and compiled \
the program into an executable called \"heat12\", you can make the \"simulate\
\" function that it contains accessible to ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "."
}], "Text"],

Cell[BoxData[
    \(\(\nInstall["\<heat12\>"]\)\)], "Input"],

Cell[TextData[{
  "(A.61)  Simulate can then be invoked just as if it were built in to ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "."
}], "Text"],

Cell["simulate[1.752, 10., 100., 50., 0., 5., 5]", "Input"],

Cell["(A.63)  Similarly, we can install \"heat13\" ...", "Text"],

Cell["Install[\"heat12\"]", "Input"],

Cell["(A.64)  ... and invoke its version of \"simulate\".", "Text"],

Cell["simulate[1.752, 10., 100., 50., 0., 5., 5]", "Input"]
}, Open  ]]
},
FrontEndVersion->"Microsoft Windows 3.0",
ScreenRectangle->{{0, 1600}, {0, 1144}},
WindowSize->{488, 568},
WindowMargins->{{Automatic, 97}, {160, Automatic}}
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1731, 51, 212, 5, 67, "Subsection"],
Cell[1946, 58, 163, 5, 33, "Text"],
Cell[2112, 65, 35, 0, 30, "Input"],
Cell[2150, 67, 223, 6, 52, "Text"],
Cell[2376, 75, 41, 0, 30, "Input"],
Cell[2420, 77, 235, 6, 52, "Text"],
Cell[2658, 85, 34, 0, 30, "Input"],
Cell[2695, 87, 144, 5, 33, "Text"],
Cell[2842, 94, 48, 1, 30, "Input"],
Cell[2893, 97, 73, 0, 33, "Text"],
Cell[2969, 99, 60, 1, 50, "Input"],
Cell[3032, 102, 53, 0, 33, "Text"],
Cell[3088, 104, 42, 1, 30, "Input"],
Cell[3133, 107, 237, 6, 52, "Text"],
Cell[3373, 115, 66, 0, 30, "Input"],
Cell[3442, 117, 99, 3, 33, "Text"],
Cell[3544, 122, 68, 0, 30, "Input"],
Cell[3615, 124, 265, 7, 71, "Text"],
Cell[3883, 133, 62, 0, 30, "Input"],
Cell[3948, 135, 89, 3, 33, "Text"],
Cell[4040, 140, 65, 0, 30, "Input"],
Cell[4108, 142, 63, 0, 33, "Text"],
Cell[4174, 144, 90, 3, 48, "Input"],
Cell[4267, 149, 119, 3, 33, "Text"],
Cell[4389, 154, 32, 0, 30, "Input"],
Cell[4424, 156, 98, 3, 33, "Text"],
Cell[4525, 161, 53, 0, 30, "Input"],
Cell[4581, 163, 58, 0, 33, "Text"],
Cell[4642, 165, 51, 0, 30, "Input"],
Cell[4696, 167, 136, 3, 52, "Text"],
Cell[4835, 172, 57, 0, 30, "Input"],
Cell[4895, 174, 110, 3, 33, "Text"],
Cell[5008, 179, 91, 3, 48, "Input"],
Cell[5102, 184, 56, 0, 33, "Text"],
Cell[5161, 186, 57, 0, 30, "Input"],
Cell[5221, 188, 155, 5, 33, "Text"],
Cell[5379, 195, 66, 1, 30, "Input"],
Cell[5448, 198, 128, 3, 52, "Text"],
Cell[5579, 203, 77, 1, 30, "Input"],
Cell[5659, 206, 156, 5, 33, "Text"],
Cell[5818, 213, 130, 3, 42, "Input"],
Cell[5951, 218, 89, 3, 33, "Text"],
Cell[6043, 223, 61, 1, 33, "Input"],
Cell[6107, 226, 106, 3, 33, "Text"],
Cell[6216, 231, 74, 1, 33, "Input"],
Cell[6293, 234, 37, 0, 33, "Text"],
Cell[6333, 236, 91, 1, 40, "Input"],
Cell[6427, 239, 74, 0, 33, "Text"],
Cell[6504, 241, 66, 1, 40, "Input"],
Cell[6573, 244, 44, 0, 33, "Text"],
Cell[6620, 246, 93, 1, 30, "Input"],
Cell[6716, 249, 90, 3, 33, "Text"],
Cell[6809, 254, 141, 3, 30, "Input"],
Cell[6953, 259, 48, 0, 33, "Text"],
Cell[7004, 261, 88, 1, 42, "Input"],
Cell[7095, 264, 143, 3, 52, "Text"],
Cell[7241, 269, 61, 0, 30, "Input"],
Cell[7305, 271, 176, 4, 52, "Text"],
Cell[7484, 277, 61, 0, 30, "Input"],
Cell[7548, 279, 66, 0, 33, "Text"],
Cell[7617, 281, 37, 0, 30, "Input"],
Cell[7657, 283, 53, 0, 33, "Text"],
Cell[7713, 285, 37, 0, 30, "Input"],
Cell[7753, 287, 52, 0, 33, "Text"],
Cell[7808, 289, 37, 0, 30, "Input"],
Cell[7848, 291, 51, 0, 33, "Text"],
Cell[7902, 293, 38, 0, 30, "Input"],
Cell[7943, 295, 137, 3, 52, "Text"],
Cell[8083, 300, 54, 0, 30, "Input"],
Cell[8140, 302, 102, 3, 33, "Text"],
Cell[8245, 307, 57, 0, 30, "Input"],
Cell[8305, 309, 210, 4, 52, "Text"],
Cell[8518, 315, 61, 0, 30, "Input"],
Cell[8582, 317, 113, 3, 33, "Text"],
Cell[8698, 322, 96, 3, 48, "Input"],
Cell[8797, 327, 55, 0, 33, "Text"],
Cell[8855, 329, 103, 3, 48, "Input"],
Cell[8961, 334, 113, 3, 33, "Text"],
Cell[9077, 339, 33, 0, 30, "Input"],
Cell[9113, 341, 69, 0, 33, "Text"],
Cell[9185, 343, 33, 0, 30, "Input"],
Cell[9221, 345, 125, 3, 52, "Text"],
Cell[9349, 350, 29, 0, 30, "Input"],
Cell[9381, 352, 70, 0, 33, "Text"],
Cell[9454, 354, 36, 0, 30, "Input"],
Cell[9493, 356, 63, 0, 33, "Text"],
Cell[9559, 358, 31, 0, 30, "Input"],
Cell[9593, 360, 66, 0, 33, "Text"],
Cell[9662, 362, 41, 0, 30, "Input"],
Cell[9706, 364, 128, 3, 52, "Text"],
Cell[9837, 369, 34, 0, 30, "Input"],
Cell[9874, 371, 32, 0, 33, "Text"],
Cell[9909, 373, 30, 0, 30, "Input"],
Cell[9942, 375, 33, 0, 33, "Text"],
Cell[9978, 377, 32, 0, 30, "Input"],
Cell[10013, 379, 98, 3, 33, "Text"],
Cell[10114, 384, 37, 0, 30, "Input"],
Cell[10154, 386, 49, 0, 33, "Text"],
Cell[10206, 388, 32, 0, 30, "Input"],
Cell[10241, 390, 89, 3, 33, "Text"],
Cell[10333, 395, 32, 0, 30, "Input"],
Cell[10368, 397, 93, 3, 33, "Text"],
Cell[10464, 402, 32, 0, 30, "Input"],
Cell[10499, 404, 108, 3, 33, "Text"],
Cell[10610, 409, 37, 0, 30, "Input"],
Cell[10650, 411, 187, 4, 52, "Text"],
Cell[10840, 417, 40, 0, 30, "Input"],
Cell[10883, 419, 75, 0, 33, "Text"],
Cell[10961, 421, 46, 0, 30, "Input"],
Cell[11010, 423, 94, 3, 33, "Text"],
Cell[11107, 428, 29, 0, 30, "Input"],
Cell[11139, 430, 35, 0, 33, "Text"],
Cell[11177, 432, 27, 0, 30, "Input"],
Cell[11207, 434, 65, 0, 33, "Text"],
Cell[11275, 436, 27, 0, 30, "Input"],
Cell[11305, 438, 62, 0, 33, "Text"],
Cell[11370, 440, 34, 0, 30, "Input"],
Cell[11407, 442, 60, 0, 33, "Text"],
Cell[11470, 444, 34, 0, 30, "Input"],
Cell[11507, 446, 137, 3, 52, "Text"],
Cell[11647, 451, 36, 0, 30, "Input"],
Cell[11686, 453, 57, 0, 33, "Text"],
Cell[11746, 455, 54, 0, 30, "Input"],
Cell[11803, 457, 289, 7, 71, "Text"],
Cell[12095, 466, 60, 1, 50, "Input"],
Cell[12158, 469, 159, 5, 33, "Text"],
Cell[12320, 476, 59, 0, 30, "Input"],
Cell[12382, 478, 64, 0, 33, "Text"],
Cell[12449, 480, 36, 0, 30, "Input"],
Cell[12488, 482, 67, 0, 33, "Text"],
Cell[12558, 484, 59, 0, 30, "Input"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

